/*
 * GTX - GLib Test Extensions
 * Copyright (C) Canonical Ltd 2010
 * Author: mikkel.kamstrup@canonical.com
 *
 * GTX is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * GTX is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with GTX.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

/*
 * This is a sample test case where a parent runs a main loop while waiting
 * for a fork()ed child process to exit, by using the gtx_wait_for_fork()
 * method.
 */

#include <unistd.h> /* For sleep() */
#include <gtx.h>

static gboolean
parent_print_hi (gpointer data)
{
  g_debug ("Parent process says hi!");
  return TRUE;
}

static gboolean
child1_print_hi (gpointer data)
{
  g_debug ("Child 1 process says hi!");
  return TRUE;
}

/* The 'main' function for the first child process.
 * It simply prints some messages for 500ms */
static gboolean
forked_child1 (gpointer data)
{  
  GMainLoop *loop;
  
  loop = g_main_loop_new (NULL, FALSE);
  g_timeout_add (100, (GSourceFunc)child1_print_hi, NULL);
  g_timeout_add (500, (GSourceFunc)gtx_quit_main_loop, loop);
  
  g_main_loop_run (loop);
  g_main_loop_unref (loop);
  
  return TRUE;
}

/* The 'main' function for the second child process.
 * It simply blocks for 100ms  */
static gboolean
forked_child2 (gpointer data)
{  
  g_debug ("Child 2 process says hi");
  gtx_yield_main_loop (100);
  g_debug ("Child 2 process done");
  
  return TRUE;
}

/* The 'main' function for the third child process.
 * It simply blocks for 10000ms (10s), which should cause a timeout */
static gboolean
forked_child3 (gpointer data)
{  
  g_debug ("Child 3 process says hi");
  gtx_yield_main_loop (10000);
  g_debug ("Child 3 process done");
  
  return TRUE;
}

/* The 'main' function for the fourth child process.
 * It simply crashes with a g_error ("Child 4 always dies. Boo hoo") */
static gboolean
forked_child4 (gpointer data)
{  
  g_error ("Child 4 always dies. Boo hoo");
  
  return TRUE;
}

#define FORK_TIMEOUT 2000

int
main (gint argc, gchar *argv[])
{
  g_type_init ();
  g_thread_init (NULL);

  /* When ever there's a main loop running the
   * parent process keep saying 'hi' */
  g_timeout_add (200, (GSourceFunc)parent_print_hi, NULL);
  
  /* Child 1 */  
  if (gtx_wait_for_fork (0, forked_child1, NULL, FORK_TIMEOUT))
    g_critical ("Child 1 process timed out");
  else
    g_debug ("Child 1 exited in due time");

  /* Assert that child 1 printed what we expected on stdout */
  g_test_trap_assert_passed ();
  g_test_trap_assert_stdout ("*Child 1 process says hi!*");

  /* Child 2 */
  if (gtx_wait_for_fork (0, forked_child2, NULL, FORK_TIMEOUT))
    g_critical ("Child 2 process timed out");
  else
    g_debug ("Child 2 exited in due time");

  /* Assert that child 2 did not crash */
  g_test_trap_assert_passed ();

  /* Child 3 will time out. It's a sluggish bastard */
  if (!gtx_wait_for_fork (0, forked_child3, NULL, FORK_TIMEOUT))
    g_debug ("Child 3 process timed out - as expected - that slacker!");
  else
    g_critical ("Child 3 did not time out as expected");

  /* Child 4 will always crash */
  if (!gtx_wait_for_fork (0, forked_child4, NULL, FORK_TIMEOUT))
    g_debug ("Child 4 process timed out - as expected - it should crash");    
  else
    g_critical ("Child 3 did not time out as expected");

  /* Assert that child 4 crashed */
  g_test_trap_assert_failed ();
  g_test_trap_assert_stderr ("*Boo hoo*");
  
  g_debug ("All tests passed");
  
  return 0;
}