/*
 * GTX - GLib Test Extensions
 * Copyright (C) Mikkel Kamstrup Erlandsen 2008 <mikkel.kamstrup@gmail.com>
 *               Scott Asofyet 2008 (wait_for_signal code)
 *               Canonical Ltd 2010               
 *
 * GTX is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * GTX is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with GTX.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <stdlib.h> /* For exit() function */
#include <unistd.h> /* For usleep() function */
#include <gobject/gvaluecollector.h>
#include <string.h> /* memset(), from G_VALUE_COLLECT() */
#include "gtx.h"

typedef struct {
    GClosure closure;
    GValue * return_value;
    GValueArray * param_values;
    GMainLoop * loop;
} WaitForSignalClosure;


gboolean
gtx_dispatch_test_case (GtxTestContext *ctx)
{
	g_assert (ctx != NULL);
	ctx->test_func (ctx->fixture, ctx->test_data);
	return FALSE;
}

void
gtx_flush_sources (gboolean may_block)
{
	GMainContext *ctx;
	
	ctx = g_main_context_default ();
	
	while (g_main_context_pending (ctx))
		g_main_context_iteration (ctx, may_block);
}

/**
 * gtx_quit_main_loop
 * @loop: The main loop to quit
 *
 * This is a convenience to close a #GMainLoop via a #GSourceFunc.
 *
 * Returns: Always %FALSE
 */
gboolean
gtx_quit_main_loop (GMainLoop *loop)
{
	g_assert (loop != NULL);
	
	if (g_main_loop_is_running (loop))
		g_main_loop_quit (loop);
	else
		g_warning ("Tried to quit non-running GMainLoop@%p", loop);
	
	return FALSE;
}

/**
 * gtx_yield_main_loop
 * @millis: Number of milli seconds to return control to the main loop
 *
 * Return control to the main loop for a specified amount of time. This
 * is to allow async test to process events in a blocking fashion.
 */
void
gtx_yield_main_loop (guint millis)
{
	GMainLoop *recursive_main;
	guint      timeout_id;
	
	recursive_main = g_main_loop_new (NULL, FALSE);
	
	timeout_id = g_timeout_add (millis,
	                            (GSourceFunc)gtx_quit_main_loop,
	                            recursive_main);
	g_main_loop_run (recursive_main);
	
	gtx_flush_sources (FALSE);

	g_source_remove (timeout_id);
	g_main_loop_unref (recursive_main);
}

static void
wait_for_signal_closure_marshal (GClosure *closure,
                                 GValue *return_value,
                                 guint n_param_values,
                                 const GValue *param_values,
                                 gpointer invocation_hint,
                                 gpointer marshal_data)
{	
    WaitForSignalClosure * wfsclosure = (WaitForSignalClosure *) closure;
    guint i;
	
    wfsclosure->param_values = g_value_array_new (n_param_values);
    for (i = 0 ; i < n_param_values ; i++)
        g_value_array_append (wfsclosure->param_values, param_values + i);

    if (return_value)
        g_value_copy (wfsclosure->return_value, return_value);

    gtx_quit_main_loop (wfsclosure->loop);

    (void) invocation_hint;
    (void) marshal_data;
}

static gboolean
wait_for_signal_values (GObject * object,
                        gint max_wait_ms,
                        guint signal_id,
                        GQuark detail,
                        GValue * return_value,
                        GValueArray ** param_values)
{
    WaitForSignalClosure * wfs;
    guint handler_id;
    guint max_timeout_id;
    gboolean timed_out = FALSE;

    wfs = (WaitForSignalClosure *)
        g_closure_new_simple (sizeof (WaitForSignalClosure), NULL);
    wfs->loop = g_main_loop_new (NULL, FALSE);
    wfs->return_value = return_value;
    wfs->param_values = NULL;

    g_closure_set_marshal (&wfs->closure, wait_for_signal_closure_marshal);

    handler_id = g_signal_connect_closure_by_id (object, signal_id, detail,
                                                 &wfs->closure, FALSE);

    max_timeout_id = g_timeout_add (max_wait_ms, (GSourceFunc)gtx_quit_main_loop,
									wfs->loop);

    g_main_loop_run (wfs->loop);
	
    g_closure_invalidate (&wfs->closure);

    if (wfs->param_values) {
        if (param_values)
            *param_values = wfs->param_values;
        else
            g_value_array_free (wfs->param_values);

        wfs->param_values = NULL;

    } else {
        timed_out = TRUE;
    }
	
    g_main_loop_unref (wfs->loop);
    wfs->loop = NULL;
	
    /* Closure will be destroyed here */
    g_signal_handler_disconnect (object, handler_id);
	g_source_remove (max_timeout_id);
	
    return timed_out;
}


/**
 * gtx_wait_for_signal
 * @object:  The object
 * @max_wait_ms:  Maximum number of milliseconds to wait before giving up
 *                and failing; passed directly to g_timeout_add().
 * @detailed_signal:  Detailed signal name for which to wait.
 * @...:  If @detailed_signal has a return value, the first vararg should
 *        be the value to return from the signal handler.  The rest of the
 *        varargs should be pointers to variables in which to store the
 *        parameters passed to the signal.  Pass NULL for any param in
 *        which you are not interested.  The caller is responsible for
 *        freeing or unreffing any strings or objects returned here.
 *
 * returns: %FALSE if the signal was received, %TRUE if the timeout was
 *          reached without the signal firing.
 */
gboolean
gtx_wait_for_signal (GObject * object,
                 gint max_wait_ms,
                 const gchar * detailed_signal,
                 ...)
{
    GSignalQuery query;
    GValue * return_value = NULL, the_return_value = { 0, };
    GValueArray * param_values = NULL;
    guint signal_id;
    GQuark detail;
    gboolean ret;
    va_list ap;

    if (! g_signal_parse_name (detailed_signal, G_OBJECT_TYPE (object),
                               &signal_id, &detail, FALSE))
        g_error ("Signal %s is invalid for object type %s",
                 detailed_signal, g_type_name (G_OBJECT_TYPE (object)));

    g_signal_query (signal_id, &query);

    va_start (ap, detailed_signal);

    if (query.return_type != G_TYPE_NONE) {
        GType t = query.return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE;
        gboolean static_scope = t & G_SIGNAL_TYPE_STATIC_SCOPE;
        gchar * error;
        g_value_init (&the_return_value, t);
        return_value = &the_return_value;
        G_VALUE_COLLECT (return_value, ap,
                         static_scope ? G_VALUE_NOCOPY_CONTENTS : 0,
                         &error);
        if (error)
            g_error ("%s: %s", G_STRLOC, error);
    }

    ret = wait_for_signal_values (object,
                                  max_wait_ms,
                                  signal_id,
                                  detail,
                                  return_value,
                                  &param_values);

    if (param_values) {
        guint i;

        /* Skip the instance */
        for (i = 1 ; i < param_values->n_values ; i++) {
            GValue * v = g_value_array_get_nth (param_values, i);
            switch (G_TYPE_FUNDAMENTAL (G_VALUE_TYPE (v))) {
                case G_TYPE_INVALID:
                case G_TYPE_NONE:
                case G_TYPE_INTERFACE:
                    break;

                case G_TYPE_CHAR:
                    {
                        gchar * p = va_arg (ap, gchar *);
                        if (p)
                            *p = g_value_get_char (v);
                    }
                    break;
                case G_TYPE_UCHAR:
                    {
                        guchar * p = va_arg (ap, guchar *);
                        if (p)
                            *p = g_value_get_uchar (v);
                    }
                    break;
                case G_TYPE_BOOLEAN:
                    {
                        gboolean * p = va_arg (ap, gboolean *);
                        if (p)
                            *p = g_value_get_boolean (v);
                    }
                    break;
                case G_TYPE_INT:
                    {
                        gint * p = va_arg (ap, gint *);
                        if (p)
                            *p = g_value_get_int (v);
                    }
                    break;
                case G_TYPE_UINT:
                    {
                        guint * p = va_arg (ap, guint *);
                        if (p)
                            *p = g_value_get_uint (v);
                    }
                    break;
                case G_TYPE_LONG:
                    {
                        glong * p = va_arg (ap, glong *);
                        if (p)
                            *p = g_value_get_long (v);
                    }
                    break;
                case G_TYPE_ULONG:
                    {
                        gulong * p = va_arg (ap, gulong *);
                        if (p)
                            *p = g_value_get_ulong (v);
                    }
                    break;
                case G_TYPE_INT64:
                    {
                        gint64 * p = va_arg (ap, gint64 *);
                        if (p)
                            *p = g_value_get_int64 (v);
                    }
                    break;
                case G_TYPE_UINT64:
                    {
                        guint64 * p = va_arg (ap, guint64 *);
                        if (p)
                            *p = g_value_get_uint64 (v);
                    }
                    break;
                case G_TYPE_ENUM:
                    {
                        gint * p = va_arg (ap, gint *);
                        if (p)
                            *p = g_value_get_int (v);
                    }
                    break;
                case G_TYPE_FLAGS:
                    {
                        guint * p = va_arg (ap, guint *);
                        if (p)
                            *p = g_value_get_uint (v);
                    }
                    break;
                case G_TYPE_FLOAT:
                    {
                        gfloat * p = va_arg (ap, gfloat *);
                        if (p)
                            *p = g_value_get_float (v);
                    }
                    break;
                case G_TYPE_DOUBLE:
                    {
                        gdouble * p = va_arg (ap, gdouble *);
                        if (p)
                            *p = g_value_get_double (v);
                    }
                    break;
                case G_TYPE_STRING:
                    {
                        gchar ** p = va_arg (ap, gchar **);
                        if (p)
                            *p = g_value_dup_string (v);
                    }
                    break;
                case G_TYPE_POINTER:
                    {
                        gpointer * p = va_arg (ap, gpointer *);
                        if (p)
                            *p = g_value_get_pointer (v);
                    }
                    break;
                case G_TYPE_BOXED:
                    {
                        gpointer * p = va_arg (ap, gpointer *);
                        if (p)
                            *p = g_value_dup_boxed (v);
                    }
                    break;
                case G_TYPE_PARAM:
                    {
                        GParamSpec ** p = va_arg (ap, GParamSpec **);
                        if (p)
                            *p = g_value_dup_param (v);
                    }
                    break;
                case G_TYPE_OBJECT:
                    {
                        GObject ** p = va_arg (ap, GObject **);
                        if (p)
                            *p = g_value_dup_object (v);
                    }
                    break;
            }
        }

        g_value_array_free (param_values);
    }

    va_end (ap);

    return ret;
}

typedef struct
{
  gint           is_alive;
  gint           time_left;
  GSourceFunc    func;
  gpointer       data;
  GMainLoop     *loop;
  GThread       *thread;
  GTestTrapFlags flags;
} GtxThreadContext;

#define TIMEOUT_TICK 100

static gboolean
_check_thread_timeout (GtxThreadContext *ctx)
{
  gint is_alive;  
  gint time_left;
  
  is_alive = g_atomic_int_get (&(ctx->is_alive));  
  if (is_alive == 0)
    {
      /* Thread finished nicely */
      g_idle_add ((GSourceFunc)gtx_quit_main_loop, ctx->loop);
      return FALSE; 
    }

  time_left = g_atomic_int_get (&(ctx->time_left)) - TIMEOUT_TICK;
  if (time_left < 0)
    {
      /* Thread timed out */
      g_idle_add ((GSourceFunc)gtx_quit_main_loop, ctx->loop);
      return FALSE;
    }

  g_atomic_int_set (&(ctx->time_left), time_left);
  return TRUE;
}

static gpointer
_dispatch_func (GtxThreadContext *ctx)
{
  ctx->func (ctx->data);
  
  g_atomic_int_set (&(ctx->is_alive), 0);
  return NULL;
}

/**
 * gtx_wait_for_thread
 * @func: A function to execute in a thread
 * @data: Argument passed to @func
 * @max_wait_ms: If the thread has not exited within this number of
 *               milliseconds it will be regarded as timed out
 * 
 * Wait for a thread while keeping a main loop running in the parent
 * process. This is different to g_test_trap_fork () which blocks the current
 * thread while waiting.
 *
 * Returns: %TRUE if the thread timed out. %FALSE if it exits before @max_wait_ms
 *          milliseconds has passed
 */
gboolean
gtx_wait_for_thread (GSourceFunc func,
                     gpointer data,
                     gint max_wait_ms)
{
  GtxThreadContext *ctx;
  GError *error;
  gboolean timed_out = FALSE;
  guint timeout_id;

  g_return_val_if_fail (func != NULL, TRUE);
  g_return_val_if_fail (max_wait_ms > 0, TRUE);
  if (!g_thread_get_initialized ())
    {
      g_critical ("g_thread_init() has not been called. Please make sure that "
                  "you do this as the first thing in your program");
      g_thread_init (NULL);
    }
  
  ctx = g_new0 (GtxThreadContext, 1);
  ctx->is_alive = 1;
  ctx->time_left = max_wait_ms;
  ctx->func = func;
  ctx->data = data;  

  /* IMPORTANT: This method is invoked from gtx_wait_for_fork() which means we
   *            are mixing threads and fork() here. That's a very bad idea
   *            in general. In order for this to not block we need to start
   *            the thread and dispatch the fork() _before_ we create the
   *            GMainLoop for the main thread. Otherwise the child process
   *            will block when starting a GMainLoop */
  error = NULL;
  ctx->thread = g_thread_create ((GThreadFunc) _dispatch_func, ctx,
                                 FALSE, &error);

  /* Race warning: We sleep 10,000 micro seconds (10ms) and assume that the
   *               thread have forked the child by then */
  usleep (10000);
  
  if (error != NULL)
    {
      g_critical ("Failed to spawn thread: %s", error->message);
      g_error_free (error);
      return TRUE;
    }

  ctx->loop = g_main_loop_new (NULL, FALSE);
  timeout_id = g_timeout_add (TIMEOUT_TICK,
                              (GSourceFunc) _check_thread_timeout,
                              ctx);
  g_main_loop_run (ctx->loop);

  timed_out = g_atomic_int_get (&(ctx->is_alive)) == 1;

  g_source_remove (timeout_id);
  g_main_loop_unref (ctx->loop);
  g_free (ctx);
  return timed_out;
}

static gboolean
_dispatch_fork (GtxThreadContext *ctx)
{
  g_return_val_if_fail (ctx->func != NULL, FALSE);

  /* g_test_trap_fork() uses micro senconds, not milliseconds */
  if (g_test_trap_fork (ctx->time_left*1000, ctx->flags))
    {
      /* Child */
      ctx->func (ctx->data);
      exit (0);
    }
  
  return TRUE;
}


/**
 * gtx_wait_for_fork
 * @test_trap_flags: A set of %GTestTrapFlags controlling how the streams of
 *                   the child are handled
 * @child_func: A function to execute in the forked child process
 * @child_data: Argument passed to @child_func
 * @max_wait_ms: If the child process has not exited within this number of
 *               milliseconds it will be regarded as timed out
 * 
 * Wait for a child process while keeping a main loop running in the parent
 * process. This is different to g_test_trap_fork () which blocks the current
 * thread while waiting.
 *
 * This function is useful for implemening tests for client/server interactions
 * such as DBus or other messaging/RPC frameworks.
 *
 * Note that this function works well with the %g_test_trap_assert*() family
 * of functions. This means that you can check the exit status of the child with
 * g_test_trap_assert_passed() and validate any output with
 * g_test_trap_assert_stdout() etc.
 *
 * Returns: %TRUE if the child timed out. %FALSE if it exits before @max_wait_ms
 *          milliseconds has passed
 *
 */
gboolean
gtx_wait_for_fork (GTestTrapFlags test_trap_flags,
                   GSourceFunc child_func,
                   gpointer child_data,
                   gint max_wait_ms)
{
  GtxThreadContext *ctx;
  gboolean result;
  
  g_return_val_if_fail (child_func != NULL, TRUE);
  g_return_val_if_fail (max_wait_ms > 0, TRUE);
  
  ctx = g_new0 (GtxThreadContext, 1);
  ctx->func = child_func;
  ctx->data = child_data;
  ctx->is_alive = 1;
  ctx->time_left = max_wait_ms;
  ctx->flags = test_trap_flags;
  
  result = gtx_wait_for_thread ((GSourceFunc) _dispatch_fork, ctx, max_wait_ms);

  g_free (ctx);
  return result;
}



static gint _last_command_status = 0;

static void
_child_watch (GPid pid,
              gint status,
              GtxThreadContext *ctx)
{
  _last_command_status = status;
  ctx->is_alive = 0;
  g_spawn_close_pid (pid);
  g_main_loop_quit (ctx->loop);
}

/**
 * gtx_assert_last_command_status
 * @status: Assert that the last command spawned via gtx_wait_for_command() had
 *          an exit status of @status
 *
 * Invoke this command following a gtx_wait_for_command() to assert that
 * the spawned command exited with the expected value
 */
void
gtx_assert_last_command_status (gint status)
{
  g_assert_cmpint (_last_command_status, ==, status);
}

/**
 * gtx_wait_for_command
 * @working_directory: The working directory to execute the command in
 * @argv: The command line to spawn
 * @max_wait_ms: After this number of milliseconds the command will be timed out
 *
 * Returns: %TRUE if the command timed out. %FALSE if it exited in due time.
 *          When this command returns %FALSE you can check the exit status of
 *          the command by calling gtx_assert_last_command_status().
 */
gboolean
gtx_wait_for_command (const gchar *working_directory,
                      gchar **argv,
                      gint max_wait_ms)
{
  GPid pid;
  GError *error;
  GtxThreadContext *ctx;
  gboolean timed_out;
  guint    timeout_id;
  
  error = NULL;
  g_spawn_async (working_directory,
                 argv,
                 NULL,
                 G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_SEARCH_PATH,
                 NULL, NULL,
                 &pid,
                 &error);
  if (error)
    {
      g_critical ("Failed to spawn child: %s", error->message);
      g_error_free (error);
      return TRUE;
    }

  ctx = g_new0 (GtxThreadContext, 1);
  ctx->is_alive = 1;
  ctx->time_left = max_wait_ms;
  ctx->loop = g_main_loop_new (NULL, FALSE);
  
  g_child_watch_add (pid, _child_watch, ctx);
  timeout_id = g_timeout_add (TIMEOUT_TICK,
                              (GSourceFunc) _check_thread_timeout,
                              ctx);
  
  /* Wait for the child */
  g_main_loop_run (ctx->loop);
  timed_out = ctx->is_alive == 1;

  g_source_remove (timeout_id);
  g_main_loop_unref (ctx->loop);
  g_free (ctx);

  return timed_out;
}